from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Chart(_message.Message):
    __slots__ = ["end", "requestid", "resolution", "series", "start"]
    class Interval(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Request(_message.Message):
        __slots__ = ["culture", "override", "period", "requestid", "resolution", "series", "tz"]
        class OverrideEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        CULTURE_FIELD_NUMBER: _ClassVar[int]
        OVERRIDE_FIELD_NUMBER: _ClassVar[int]
        PERIOD_FIELD_NUMBER: _ClassVar[int]
        REQUESTID_FIELD_NUMBER: _ClassVar[int]
        RESOLUTION_FIELD_NUMBER: _ClassVar[int]
        SERIES_FIELD_NUMBER: _ClassVar[int]
        TZ_FIELD_NUMBER: _ClassVar[int]
        culture: str
        override: _containers.ScalarMap[str, str]
        period: Chart.Interval
        requestid: str
        resolution: Chart.Interval
        series: _containers.RepeatedScalarFieldContainer[str]
        tz: str
        def __init__(self, requestid: _Optional[str] = ..., resolution: _Optional[_Union[Chart.Interval, str]] = ..., culture: _Optional[str] = ..., period: _Optional[_Union[Chart.Interval, str]] = ..., series: _Optional[_Iterable[str]] = ..., tz: _Optional[str] = ..., override: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class Serie(_message.Message):
        __slots__ = ["data", "expires", "id", "times", "type"]
        DATA_FIELD_NUMBER: _ClassVar[int]
        EXPIRES_FIELD_NUMBER: _ClassVar[int]
        ID_FIELD_NUMBER: _ClassVar[int]
        TIMES_FIELD_NUMBER: _ClassVar[int]
        TYPE_FIELD_NUMBER: _ClassVar[int]
        data: _struct_pb2.ListValue
        expires: str
        id: str
        times: str
        type: str
        def __init__(self, times: _Optional[str] = ..., expires: _Optional[str] = ..., data: _Optional[_Union[_struct_pb2.ListValue, _Mapping]] = ..., id: _Optional[str] = ..., type: _Optional[str] = ...) -> None: ...
    END_FIELD_NUMBER: _ClassVar[int]
    P10Y: Chart.Interval
    P1D: Chart.Interval
    P1M: Chart.Interval
    P1W: Chart.Interval
    P1Y: Chart.Interval
    P3M: Chart.Interval
    P3Y: Chart.Interval
    P50Y: Chart.Interval
    P5Y: Chart.Interval
    P6M: Chart.Interval
    PT15M: Chart.Interval
    PT15S: Chart.Interval
    PT1H: Chart.Interval
    PT1M: Chart.Interval
    PT1S: Chart.Interval
    PT30M: Chart.Interval
    PT30S: Chart.Interval
    PT5M: Chart.Interval
    PT60M: Chart.Interval
    REQUESTID_FIELD_NUMBER: _ClassVar[int]
    RESOLUTION_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    YTD: Chart.Interval
    end: str
    requestid: str
    resolution: str
    series: _containers.RepeatedCompositeFieldContainer[Chart.Serie]
    start: str
    def __init__(self, requestid: _Optional[str] = ..., start: _Optional[str] = ..., end: _Optional[str] = ..., resolution: _Optional[str] = ..., series: _Optional[_Iterable[_Union[Chart.Serie, _Mapping]]] = ...) -> None: ...

class Metadata(_message.Message):
    __slots__ = ["request_duration", "response_datetime"]
    REQUEST_DURATION_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_DATETIME_FIELD_NUMBER: _ClassVar[int]
    request_duration: _duration_pb2.Duration
    response_datetime: _timestamp_pb2.Timestamp
    def __init__(self, response_datetime: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., request_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class Quotecast(_message.Message):
    __slots__ = ["json_data", "metadata"]
    class Request(_message.Message):
        __slots__ = ["subscriptions", "unsubscriptions"]
        class SubscriptionsEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: _struct_pb2.ListValue
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.ListValue, _Mapping]] = ...) -> None: ...
        class UnsubscriptionsEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: _struct_pb2.ListValue
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.ListValue, _Mapping]] = ...) -> None: ...
        SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
        UNSUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
        subscriptions: _containers.MessageMap[str, _struct_pb2.ListValue]
        unsubscriptions: _containers.MessageMap[str, _struct_pb2.ListValue]
        def __init__(self, subscriptions: _Optional[_Mapping[str, _struct_pb2.ListValue]] = ..., unsubscriptions: _Optional[_Mapping[str, _struct_pb2.ListValue]] = ...) -> None: ...
    JSON_DATA_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    json_data: str
    metadata: Metadata
    def __init__(self, json_data: _Optional[str] = ..., metadata: _Optional[_Union[Metadata, _Mapping]] = ...) -> None: ...

class Ticker(_message.Message):
    __slots__ = ["metadata", "product_list", "products"]
    class Metrics(_message.Message):
        __slots__ = ["metrics"]
        class MetricsEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: float
            def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
        METRICS_FIELD_NUMBER: _ClassVar[int]
        metrics: _containers.ScalarMap[str, float]
        def __init__(self, metrics: _Optional[_Mapping[str, float]] = ...) -> None: ...
    class ProductsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Ticker.Metrics
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Ticker.Metrics, _Mapping]] = ...) -> None: ...
    METADATA_FIELD_NUMBER: _ClassVar[int]
    PRODUCTS_FIELD_NUMBER: _ClassVar[int]
    PRODUCT_LIST_FIELD_NUMBER: _ClassVar[int]
    metadata: Metadata
    product_list: _containers.RepeatedScalarFieldContainer[str]
    products: _containers.MessageMap[str, Ticker.Metrics]
    def __init__(self, metadata: _Optional[_Union[Metadata, _Mapping]] = ..., products: _Optional[_Mapping[str, Ticker.Metrics]] = ..., product_list: _Optional[_Iterable[str]] = ...) -> None: ...



# IMPORTATION STANDARD
import logging
from datetime import datetime
from typing import Optional

# IMPORTATION THIRD PARTY
import json
import pandas as pd
import requests
from google.protobuf import json_format
from google.protobuf.message import Message

# IMPORTATION INTERNAL
from degiro_connector.core.constants import urls
from degiro_connector.quotecast.models.quotecast_pb2 import (
    Chart,
)
from degiro_connector.core.abstracts.abstract_action import AbstractAction


class ChartHelper:
    UNITS_MATCHING = {
        "PT1S": 1,
        "PT1M": 60,
        "PT1H": 3600,
        "P1D": 3600 * 24,
        "P1W": 3600 * 24 * 7,
        "P1M": 3600 * 24 * 30,
        "P1Y": 3600 * 24 * 365,  # Year resolution doesn't seem to be supported
    }

    @staticmethod
    def parse_start_timestamp(times: str) -> float:
        """Extract the start timestamp of a timeserie.
        Args:
            times (str):
                Combination of `start date` and `resolution` of the serie.
                Example :
                    times = "2021-10-28/P6M"
                    times = "2021-11-03T00:00:00/PT1H"
        Returns:
            float:
                Timestamp of the start date of the serie.
        """

        (start, resolution) = times.rsplit(sep="/", maxsplit=1)

        date_format = ""

        # with "PT1M":
        # {'requestid': '', 'start': '2023-06-30T00:00:00+02:00', 'end': '2023-06-30T21:59:59+02:00', 'resolution': 'PT1M', 'series': [{'times': '2023-06-30/PT1M', 'expires': '2023-07-02T14:10:43.1908462+02:00', 'data':
        # error when doing : ChartHelper.format_chart(chart=chart, copy=False) :
        # time data '2023-06-30' does not match format '%Y-%m-%dT%H:%M:%S'
        # same with "PT5M"
        # {'requestid': '', 'start': '2023-06-30T00:00:00+02:00', 'end': '2023-06-30T21:59:59+02:00', 'resolution': 'PT5M', 'series': [{'times': '2023-06-30/PT5M', 'expires': '2023-07-02T14:17:36.4487987+02:00', 'data':
        # same with "PT{xxx}" ...
        # conclusion : it's always --> date_format = "%Y-%m-%d"

        """
        previous code:
        if resolution.startswith("PT"):
            date_format = "%Y-%m-%dT%H:%M:%S"
        else:
            date_format = "%Y-%m-%d"
        """

        # new code :
        date_format = "%Y-%m-%d"

        start_datetime = datetime.strptime(start, date_format)
        start_timestamp = start_datetime.timestamp()

        return start_timestamp

    @classmethod
    def parse_interval_in_seconds(cls, times: str) -> int:
        """Extract the interval of a timeserie.
        Args:
            times (str):
                Combination of `start date` and `resolution` of the serie.
                Example :
                    times = "2021-10-28/P6M"
                    times = "2021-11-03T00:00:00/PT1H"
        Raises:
            AttributeError:
                if the resolution is unknown.
        Returns:
            int:
                Number of seconds in the interval.
        """
        (_start, resolution) = times.rsplit(sep="/", maxsplit=1)

        prefix = ""
        if resolution.startswith("PT"):
            prefix = "PT"
        elif resolution.startswith("P"):
            prefix = "P"
        else:
            raise AttributeError("Unkown resolution")

        unit = prefix + "1" + resolution[-1]
        number = int(resolution[len(prefix) : -1])

        interval = cls.UNITS_MATCHING[unit] * int(number)

        return interval

    @classmethod
    def format_serie(cls, serie: Chart.Serie, copy: bool = True) -> Chart.Serie:
        """By default a time serie uses the order as index.
        This method convert the indexes into `timestamp in seconds`.
        Args:
            serie (Chart.Serie):
                Serie to format.
            copy (bool, optional):
                Whether or not to make a copy before the formatting.
                Defaults to True.

        Returns:
            Chart.Serie: [description]
        """
        if copy:
            serie_copy = Chart.Serie()
            serie_copy.CopyFrom(serie)
            serie = serie_copy

        if serie.type in ["time", "ohlc"]:
            times = serie.times
            start = cls.parse_start_timestamp(times=times)
            interval = cls.parse_interval_in_seconds(times=times)

            for datapoint in serie.data:
                datapoint[0] = start + datapoint[0] * interval

        return serie

    @classmethod
    def format_chart(cls, chart: Chart, copy: bool = True) -> Chart:
        """By default a time serie uses the order as index.
        This method convert series`s indexes into `timestamp in seconds`.
        Args:
            chart (Chart):
                Chart containing the series.
            copy (bool, optional):
                Whether or not to make a copy before the formatting.
                Defaults to True.
        Returns:
            Chart:
                Formatted Chart.
        """

        if copy:
            chart_copy = Chart()
            chart_copy.CopyFrom(chart)
            chart = chart_copy

        for serie in chart.series:
            cls.format_serie(serie=serie, copy=False)

        return chart

    @staticmethod
    def message_to_dict(message: Message) -> dict:
        return json_format.MessageToDict(
            message=message,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
            descriptor_pool=None,
            float_precision=None,
        )

    @classmethod
    def serie_to_df(cls, serie: Chart.Serie) -> pd.DataFrame:
        """Converts a timeserie into a DataFrame.
        Only series with the following types can be converted into DataFrame :
        - serie.type == "time"
        - serie.type == "ohlc"
        Beware of series with the following type :
         - serie.type == "object"
        These are not actual timeseries and can't converted into DataFrame.
        Args:
            serie (Chart.Serie):
                The serie to convert.
        Raises:
            AttributeError:
                If the serie.type is incorrect.
        Returns:
            pd.DataFrame: [description]
        """
        columns = []
        if serie.type == "ohlc" and serie.id.startswith("ohlc:"):
            columns = [
                "timestamp",
                "open",
                "high",
                "low",
                "close",
            ]
        elif serie.type == "time" and serie.id.startswith("price:"):
            columns = [
                "timestamp",
                "price",
            ]
        elif serie.type == "time" and serie.id.startswith("volume:"):
            columns = [
                "timestamp",
                "volume",
            ]
        elif serie.type == "object":
            raise AttributeError(f"Not a timeserie, serie.type = {serie.type}")
        else:
            raise AttributeError(f"Unknown serie, serie.type = {serie.type}")

        return pd.DataFrame.from_records(serie.data, columns=columns)


class ActionGetChart(AbstractAction):
    CALLBACK = "vwd.hchart.seriesRequestManager.sync_response"

    @staticmethod
    def chart_request_to_api(
        request: Chart.Request,
        user_token: int,
    ) -> dict:
        request_dict = json_format.MessageToDict(
            message=request,
            including_default_value_fields=True,
            preserving_proto_field_name=False,
            use_integers_for_enums=False,
            descriptor_pool=None,
            float_precision=None,
        )
        request_dict["format"] = "json"
        request_dict["callback"] = "vwd.hchart.seriesRequestManager.sync_response"
        request_dict["userToken"] = user_token

        for key, value in request.override.items():
            request_dict[key] = value

        return request_dict

    @staticmethod
    def api_to_chart(payload: dict) -> Chart:
        for serie in payload["series"]:
            if serie["type"] == "object":
                serie["data"] = [serie["data"]]

        chart = Chart()
        json_format.ParseDict(
            js_dict=payload,
            message=chart,
            ignore_unknown_fields=True,
            descriptor_pool=None,
        )

        return chart

    @classmethod
    def get_chart(
        cls,
        request: Chart.Request,
        user_token: int,
        logger: logging.Logger | None = None,
        raw: bool = False,
        session: requests.Session | None = None,
    ) -> Optional[Chart]:
        """Fetches chart's data.
        Args:
            request (Chart.Request):
                Example :
                    request = Chart.Request()
                    request.culture = "fr-FR"
                    request.period = Chart.Interval.PT1H
                    request.requestid = "1"
                    request.resolution = Chart.Interval.P1D
                    # request.series.append("issueid:360148977")
                    request.series.append("price:issueid:360148977")
                    # request.series.append("ohlc:issueid:360148977")
                    # request.series.append("volume:issueid:360148977")
                    # request.series.append("vwdkey:AAPL.BATS,E")
                    # request.series.append("price:vwdkey:AAPL.BATS,E")
                    # request.series.append("ohlc:vwdkey:AAPL.BATS,E")
                    # request.series.append("volume:vwdkey:AAPL.BATS,E")
                    request.tz = "Europe/Paris"
                    request.override["resolution"] = "P1D"
                    request.override["period"] = "P1W"

                    The parameter `request.override` allows overriding
                    the request sent to Degiro's API.
            user_token (int):
                User identifier in Degiro's API.
            raw (bool, optional):
                Whether are not we want the raw API response.
                Defaults to False.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Returns:
            Chart: Data of the chart.
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        url = urls.CHART
        params = cls.chart_request_to_api(
            request=request,
            user_token=user_token,
        )

        http_request = requests.Request(method="GET", url=url, params=params)
        prepped = session.prepare_request(http_request)
        response_raw = None

        try:
            response_raw = session.send(prepped)
            response_raw.raise_for_status()
            response_dict = json.loads(response_raw.text[len(cls.CALLBACK) + 1 : -1])

            if raw is True:
                return response_dict
            else:
                return cls.api_to_chart(payload=response_dict)

        except requests.HTTPError as e:
            status_code = getattr(response_raw, "status_code", "No status_code found.")
            text = getattr(response_raw, "text", "No text found.")
            logger.fatal(status_code)
            logger.fatal(text)
            return None
        except Exception as e:
            logger.fatal(e)
            return None

    def call(
        self,
        request: Chart.Request,
        raw: bool = False,
    ) -> Optional[Chart]:
        session = self.session_storage.session
        logger = self.logger
        credentials = self.credentials
        user_token = credentials["user_token"]

        return self.get_chart(
            request=request,
            user_token=user_token,
            logger=logger,
            raw=raw,
            session=session,
        )



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from degiro_connector.trading.models import trading_pb2 as _trading_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ConfirmOrder(_message.Message):
    __slots__ = ["confirmation_id", "order"]
    CONFIRMATION_ID_FIELD_NUMBER: _ClassVar[int]
    ORDER_FIELD_NUMBER: _ClassVar[int]
    confirmation_id: _wrappers_pb2.StringValue
    order: _trading_pb2.Order
    def __init__(self, confirmation_id: _Optional[_Union[_wrappers_pb2.StringValue, _Mapping]] = ..., order: _Optional[_Union[_trading_pb2.Order, _Mapping]] = ...) -> None: ...

class ProductSearch(_message.Message):
    __slots__ = ["bonds", "etfs", "funds", "futures", "leverageds", "lookup", "options", "stocks", "warrants"]
    BONDS_FIELD_NUMBER: _ClassVar[int]
    ETFS_FIELD_NUMBER: _ClassVar[int]
    FUNDS_FIELD_NUMBER: _ClassVar[int]
    FUTURES_FIELD_NUMBER: _ClassVar[int]
    LEVERAGEDS_FIELD_NUMBER: _ClassVar[int]
    LOOKUP_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    STOCKS_FIELD_NUMBER: _ClassVar[int]
    WARRANTS_FIELD_NUMBER: _ClassVar[int]
    bonds: _trading_pb2.ProductSearch.RequestBonds
    etfs: _trading_pb2.ProductSearch.RequestETFs
    funds: _trading_pb2.ProductSearch.RequestFunds
    futures: _trading_pb2.ProductSearch.RequestFutures
    leverageds: _trading_pb2.ProductSearch.RequestLeverageds
    lookup: _trading_pb2.ProductSearch.RequestLookup
    options: _trading_pb2.ProductSearch.RequestOptions
    stocks: _trading_pb2.ProductSearch.RequestStocks
    warrants: _trading_pb2.ProductSearch.RequestWarrants
    def __init__(self, bonds: _Optional[_Union[_trading_pb2.ProductSearch.RequestBonds, _Mapping]] = ..., etfs: _Optional[_Union[_trading_pb2.ProductSearch.RequestETFs, _Mapping]] = ..., funds: _Optional[_Union[_trading_pb2.ProductSearch.RequestFunds, _Mapping]] = ..., futures: _Optional[_Union[_trading_pb2.ProductSearch.RequestFutures, _Mapping]] = ..., leverageds: _Optional[_Union[_trading_pb2.ProductSearch.RequestLeverageds, _Mapping]] = ..., lookup: _Optional[_Union[_trading_pb2.ProductSearch.RequestLookup, _Mapping]] = ..., options: _Optional[_Union[_trading_pb2.ProductSearch.RequestOptions, _Mapping]] = ..., stocks: _Optional[_Union[_trading_pb2.ProductSearch.RequestStocks, _Mapping]] = ..., warrants: _Optional[_Union[_trading_pb2.ProductSearch.RequestWarrants, _Mapping]] = ...) -> None: ...

class SetConfig(_message.Message):
    __slots__ = ["auto_connect", "credentials"]
    AUTO_CONNECT_FIELD_NUMBER: _ClassVar[int]
    CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
    auto_connect: bool
    credentials: _trading_pb2.Credentials
    def __init__(self, credentials: _Optional[_Union[_trading_pb2.Credentials, _Mapping]] = ..., auto_connect: bool = ...) -> None: ...



# IMPORTATION STANDARD
import requests
import logging
from typing import Dict, Union

# IMPORTATION THIRD PARTY

# IMPORTATION INTERNAL
from degiro_connector.core.constants import urls
from degiro_connector.trading.models.trading_pb2 import (
    Credentials,
    Order,
    Update,
)
from degiro_connector.core.abstracts.abstract_action import AbstractAction


class ActionGetUpdate(AbstractAction):
    ACTION_MATCHING = {
        "B": Order.Action.Value("BUY"),
        "S": Order.Action.Value("SELL"),
    }
    ORDER_MATCHING = {
        "buysell": "action",
        "contractSize": "contract_size",
        "contractType": "contract_type",
        "currency": "currency",
        "date": "hour",
        "id": "id",
        "isDeletable": "is_deletable",
        "isModifiable": "is_modifiable",
        "orderTimeTypeId": "time_type",
        "orderTypeId": "order_type",
        "price": "price",
        "product": "product",
        "productId": "product_id",
        "quantity": "quantity",
        "size": "size",
        "stopPrice": "stop_price",
        "totalOrderValue": "total_order_value",
    }
    UPDATE_OPTION_MATCHING = {
        Update.Option.Value("ALERTS"): "alerts",
        Update.Option.Value("CASHFUNDS"): "cashFunds",
        Update.Option.Value("HISTORICALORDERS"): "historicalOrders",
        Update.Option.Value("ORDERS"): "orders",
        Update.Option.Value("PORTFOLIO"): "portfolio",
        Update.Option.Value("TOTALPORTFOLIO"): "totalPortfolio",
        Update.Option.Value("TRANSACTIONS"): "transactions",
    }

    @classmethod
    def update_request_list_to_api(cls, request_list: Update.RequestList) -> dict:
        """Makes a payload compatible with the API.
        Parameters:
            update_option_list {UpdateOptionList}
                List of option available from grpc "consume_update".
        Returns:
            {dict}
                Payload that Degiro's update endpoint can understand.
        """

        payload = dict()

        for request in request_list.values:
            option = cls.UPDATE_OPTION_MATCHING[request.option]
            payload[option] = request.last_updated

        return payload

    @classmethod
    def setup_update_orders(cls, update: Update, payload: dict):
        """Build an "Order" object using "dict" returned by the API.
        Parameters:
            order {dict}
                Order dict straight from Degiro's API
        Returns:
            {Order}
        """

        if "orders" in payload:
            update.orders.last_updated = payload["orders"]["lastUpdated"]

            for order in payload["orders"]["value"]:
                order_dict = dict()
                for attribute in order["value"]:
                    if (
                        "name" in attribute
                        and "value" in attribute
                        and attribute["name"] in cls.ORDER_MATCHING
                    ):
                        order_dict[cls.ORDER_MATCHING[attribute["name"]]] = attribute[
                            "value"
                        ]

                order_dict["action"] = cls.ACTION_MATCHING[order_dict["action"]]
                update.orders.values.append(Order(**order_dict))

    @classmethod
    def setup_update_portfolio(cls, update: Update, payload: dict):
        if "portfolio" in payload:
            update.portfolio.last_updated = payload["portfolio"]["lastUpdated"]

            for positionrow in payload["portfolio"]["value"]:
                value = update.portfolio.values.add()
                for attribute in positionrow["value"]:
                    if "name" in attribute and "value" in attribute:
                        value[attribute["name"]] = attribute["value"]

    @classmethod
    def setup_update_total_portfolio(cls, update: Update, payload: dict):
        if "totalPortfolio" in payload:
            update.total_portfolio.last_updated = payload["totalPortfolio"][
                "lastUpdated"
            ]

            for attribute in payload["totalPortfolio"]["value"]:
                if "name" in attribute and "value" in attribute:
                    name = attribute["name"]
                    value = attribute["value"]
                    update.total_portfolio.values[name] = value

    @classmethod
    def update_to_grpc(cls, payload: dict) -> Update:
        update = Update()
        update.response_datetime.GetCurrentTime()

        # ORDERS
        cls.setup_update_orders(update=update, payload=payload)

        # PORTFOLIO
        cls.setup_update_portfolio(
            update=update,
            payload=payload,
        )

        # TOTALPORTFOLIO
        cls.setup_update_total_portfolio(
            update=update,
            payload=payload,
        )

        return update

    @classmethod
    def get_update(
        cls,
        request_list: Update.RequestList,
        session_id: str,
        credentials: Credentials,
        logger: logging.Logger = None,
        raw: bool = False,
        session: requests.Session = None,
    ) -> Union[Update, Dict, None]:
        """Retrieve information from Degiro's Trading Update endpoint.
        Args:
            request (Update.RequestList):
                List of options that we want to retrieve from the endpoint.
                Example :
                    request = Update.RequestList()
                    request.list.extend(
                        [
                            Update.Request(
                                option=Update.Option.ALERTS,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.CASHFUNDS,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.HISTORICALORDERS,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.ORDERS,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.PORTFOLIO,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.TOTALPORTFOLIO,
                                last_updated=0,
                            ),
                            Update.Request(
                                option=Update.Option.TRANSACTIONS,
                                last_updated=0,
                            ),
                        ]
                    )
            session_id (str):
                API's session id.
            credentials (Credentials):
                Credentials containing the parameter "int_account".
            raw (bool, optional):
                Whether are not we want the raw API response.
                Defaults to False.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Returns:
            Update: API response.
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        int_account = credentials.int_account
        url = urls.UPDATE
        url = f"{url}/{int_account};jsessionid={session_id}"

        params = cls.update_request_list_to_api(request_list=request_list)
        params["intAccount"] = int_account
        params["sessionId"] = session_id

        request = requests.Request(method="GET", url=url, params=params)
        prepped = session.prepare_request(request)
        response_raw = None

        try:
            response_raw = session.send(prepped)
            response_raw.raise_for_status()
            response_dict = response_raw.json()

            if raw is True:
                return response_dict
            else:
                return cls.update_to_grpc(
                    payload=response_dict,
                )
        except Exception as e:
            logger.fatal("error")
            logger.fatal(response_raw)
            logger.fatal(e)
            return None

    def call(
        self,
        request_list: Update.RequestList,
        raw: bool = False,
    ) -> Union[Update, Dict, None]:
        credentials = self.credentials
        connection_storage = self.connection_storage
        session_id = connection_storage.session_id
        session = self.session_storage.session
        logger = self.logger

        return self.get_update(
            request_list=request_list,
            session_id=session_id,
            credentials=credentials,
            logger=logger,
            raw=raw,
            session=session,
        )



QUOTECAST_TIMEOUT = 15
TRADING_TIMEOUT = 1800  # [s] This value is used in the timeout documentation, see: https://github.com/Chavithra/degiro-connector#38-is-there-a-timeout-. Whenever this constant is changed, please update the docu



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from degiro_connector.trading.models import trading_pb2 as degiro__connector_dot_trading_dot_models_dot_trading__pb2
from degiro_connector.trading.models import trading_relay_pb2 as degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


class TradingRelayStub(object):
    """SERVICES
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.confirm_order = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/confirm_order',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ConfirmOrder.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.ConfirmationResponse.FromString,
                )
        self.product_search = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/product_search',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ProductSearch.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.FromString,
                )
        self.set_config = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/set_config',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.SetConfig.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )
        self.check_order = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/check_order',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.CheckingResponse.FromString,
                )
        self.connect = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/connect',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                )
        self.delete_order = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/delete_order',
                request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )
        self.get_account_info = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_account_info',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_struct__pb2.Struct.FromString,
                )
        self.get_account_overview = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_account_overview',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.FromString,
                )
        self.get_agenda = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_agenda',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.FromString,
                )
        self.get_cash_account_report = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_cash_account_report',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.FromString,
                )
        self.get_client_details = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_client_details',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_struct__pb2.Struct.FromString,
                )
        self.get_company_profile = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_company_profile',
                request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyProfile.FromString,
                )
        self.get_company_ratios = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_company_ratios',
                request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyRatios.FromString,
                )
        self.get_config = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_config',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_struct__pb2.Struct.FromString,
                )
        self.get_favourites_list = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_favourites_list',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Favourites.FromString,
                )
        self.get_financial_statements = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_financial_statements',
                request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.FinancialStatements.FromString,
                )
        self.get_latest_news = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_latest_news',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.FromString,
                )
        self.get_news_by_company = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_news_by_company',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.FromString,
                )
        self.get_orders_history = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_orders_history',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.FromString,
                )
        self.get_products_config = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_products_config',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.Config.FromString,
                )
        self.get_products_info = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_products_info',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.FromString,
                )
        self.get_top_news_preview = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_top_news_preview',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TopNewsPreview.FromString,
                )
        self.get_transactions_history = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_transactions_history',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.FromString,
                )
        self.get_update = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/get_update',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.RequestList.SerializeToString,
                response_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.FromString,
                )
        self.logout = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/logout',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )
        self.update_order = channel.unary_unary(
                '/degiro_connector.trading_relay.TradingRelay/update_order',
                request_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )


class TradingRelayServicer(object):
    """SERVICES
    """

    def confirm_order(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def product_search(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def set_config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def check_order(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def connect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def delete_order(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_account_info(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_account_overview(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_agenda(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_cash_account_report(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_client_details(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_company_profile(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_company_ratios(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_favourites_list(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_financial_statements(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_latest_news(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_news_by_company(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_orders_history(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_products_config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_products_info(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_top_news_preview(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_transactions_history(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_update(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def logout(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def update_order(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TradingRelayServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'confirm_order': grpc.unary_unary_rpc_method_handler(
                    servicer.confirm_order,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ConfirmOrder.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.ConfirmationResponse.SerializeToString,
            ),
            'product_search': grpc.unary_unary_rpc_method_handler(
                    servicer.product_search,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ProductSearch.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.SerializeToString,
            ),
            'set_config': grpc.unary_unary_rpc_method_handler(
                    servicer.set_config,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.SetConfig.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
            'check_order': grpc.unary_unary_rpc_method_handler(
                    servicer.check_order,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.CheckingResponse.SerializeToString,
            ),
            'connect': grpc.unary_unary_rpc_method_handler(
                    servicer.connect,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            ),
            'delete_order': grpc.unary_unary_rpc_method_handler(
                    servicer.delete_order,
                    request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
            'get_account_info': grpc.unary_unary_rpc_method_handler(
                    servicer.get_account_info,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_struct__pb2.Struct.SerializeToString,
            ),
            'get_account_overview': grpc.unary_unary_rpc_method_handler(
                    servicer.get_account_overview,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.SerializeToString,
            ),
            'get_agenda': grpc.unary_unary_rpc_method_handler(
                    servicer.get_agenda,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.SerializeToString,
            ),
            'get_cash_account_report': grpc.unary_unary_rpc_method_handler(
                    servicer.get_cash_account_report,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.SerializeToString,
            ),
            'get_client_details': grpc.unary_unary_rpc_method_handler(
                    servicer.get_client_details,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_struct__pb2.Struct.SerializeToString,
            ),
            'get_company_profile': grpc.unary_unary_rpc_method_handler(
                    servicer.get_company_profile,
                    request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyProfile.SerializeToString,
            ),
            'get_company_ratios': grpc.unary_unary_rpc_method_handler(
                    servicer.get_company_ratios,
                    request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyRatios.SerializeToString,
            ),
            'get_config': grpc.unary_unary_rpc_method_handler(
                    servicer.get_config,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_struct__pb2.Struct.SerializeToString,
            ),
            'get_favourites_list': grpc.unary_unary_rpc_method_handler(
                    servicer.get_favourites_list,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Favourites.SerializeToString,
            ),
            'get_financial_statements': grpc.unary_unary_rpc_method_handler(
                    servicer.get_financial_statements,
                    request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.FinancialStatements.SerializeToString,
            ),
            'get_latest_news': grpc.unary_unary_rpc_method_handler(
                    servicer.get_latest_news,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.SerializeToString,
            ),
            'get_news_by_company': grpc.unary_unary_rpc_method_handler(
                    servicer.get_news_by_company,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.SerializeToString,
            ),
            'get_orders_history': grpc.unary_unary_rpc_method_handler(
                    servicer.get_orders_history,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.SerializeToString,
            ),
            'get_products_config': grpc.unary_unary_rpc_method_handler(
                    servicer.get_products_config,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.Config.SerializeToString,
            ),
            'get_products_info': grpc.unary_unary_rpc_method_handler(
                    servicer.get_products_info,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.SerializeToString,
            ),
            'get_top_news_preview': grpc.unary_unary_rpc_method_handler(
                    servicer.get_top_news_preview,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TopNewsPreview.SerializeToString,
            ),
            'get_transactions_history': grpc.unary_unary_rpc_method_handler(
                    servicer.get_transactions_history,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.Request.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.SerializeToString,
            ),
            'get_update': grpc.unary_unary_rpc_method_handler(
                    servicer.get_update,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.RequestList.FromString,
                    response_serializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.SerializeToString,
            ),
            'logout': grpc.unary_unary_rpc_method_handler(
                    servicer.logout,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
            'update_order': grpc.unary_unary_rpc_method_handler(
                    servicer.update_order,
                    request_deserializer=degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'degiro_connector.trading_relay.TradingRelay', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TradingRelay(object):
    """SERVICES
    """

    @staticmethod
    def confirm_order(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/confirm_order',
            degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ConfirmOrder.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.ConfirmationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def product_search(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/product_search',
            degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.ProductSearch.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def set_config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/set_config',
            degiro__connector_dot_trading_dot_models_dot_trading__relay__pb2.SetConfig.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def check_order(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/check_order',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.CheckingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def connect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/connect',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def delete_order(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/delete_order',
            google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_account_info(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_account_info',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_struct__pb2.Struct.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_account_overview(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_account_overview',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.AccountOverview.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_agenda(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_agenda',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Agenda.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_cash_account_report(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_cash_account_report',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.CashAccountReport.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_client_details(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_client_details',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_struct__pb2.Struct.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_company_profile(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_company_profile',
            google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyProfile.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_company_ratios(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_company_ratios',
            google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.CompanyRatios.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_config',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_struct__pb2.Struct.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_favourites_list(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_favourites_list',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Favourites.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_financial_statements(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_financial_statements',
            google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.FinancialStatements.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_latest_news(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_latest_news',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.LatestNews.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_news_by_company(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_news_by_company',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.NewsByCompany.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_orders_history(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_orders_history',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.OrdersHistory.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_products_config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_products_config',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductSearch.Config.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_products_info(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_products_info',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.ProductsInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_top_news_preview(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_top_news_preview',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.TopNewsPreview.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_transactions_history(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_transactions_history',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.Request.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.TransactionsHistory.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/get_update',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.RequestList.SerializeToString,
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Update.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def logout(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/logout',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def update_order(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.trading_relay.TradingRelay/update_order',
            degiro__connector_dot_trading_dot_models_dot_trading__pb2.Order.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



HEADERS = {
    "Accept": "application/json, text/plain, */*",
    # Requests library doesn't support "br" (brotli algorithm)
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7",
    # The length is setup automatically by the Requests library
    # "Content-Length": "39",
    "Content-Type": "application/json;charset=UTF-8",
    "Origin": "https://trader.degiro.nl",
    "Referer": "https://trader.degiro.nl/login/fr",
    "Sec-CH-UA": '" Not A;Brand";v="99", "Chromium";v="102", "Google Chrome";v="102"',
    "Sec-CH-UA-Mobile": "?0",
    "Sec-CH-UA-Platform": "Windows",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/102.0.5005.63 Safari/537.36",
}



# IMPORTATION STANDARD
import abc
import importlib
import importlib.util
from types import ModuleType
from typing import List, Optional

# IMPORTATION THIRDPARTY

# IMPORTATION INTERNAL

"""
What to load ?
- Quotecast : Actions
    degiro_connector/quotecast/actions/<action>/action_<action>.py

- Trading : Actions
    degiro_connector/trading/actions/<action>/action_<action>.py

"""


class Pair:
    @property
    def module_path(self) -> str:
        return self._module_path

    @property
    def class_name(self) -> str:
        return self._class_name

    def __init__(self, module_path: str, class_name: str):
        self._module_path = module_path
        self._class_name = class_name


class InitArgs:
    @property
    def args(self) -> list:
        return self._args

    @property
    def kwargs(self) -> dict:
        return self._kwargs

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs


class LazyLoader(abc.ABC):
    @staticmethod
    def load_instance(
        module: Optional[ModuleType],
        class_name: str,
        init_args: InitArgs = None,
    ) -> Optional[object]:
        """Load an instance of a class if possible.

        Args:
            module (ModuleType): [description]
            class_name (str): [description]

        Returns:
            object: [description]
        """

        instance = None
        if (
            module is not None
            and isinstance(module, ModuleType)
            and hasattr(module, class_name)
        ):
            init_args = init_args or InitArgs()
            args = init_args.args
            kwargs = init_args.kwargs
            instance = getattr(module, class_name)(*args, **kwargs)

        return instance

    @staticmethod
    def load_module(module_path: str) -> Optional[ModuleType]:
        """Load a module from a path.

        Args:
            module_path (str):
                Module"s path.

        Returns:
            Optional[ModuleType]:
                Loaded module or None.
        """

        try:
            spec = importlib.util.find_spec(module_path)
        except ModuleNotFoundError:
            spec = None

        if spec is None:
            return None
        else:
            # module = importlib.util.module_from_spec(spec)
            # sys.modules[module_path] = module
            # spec.loader.exec_module(module)
            module = importlib.import_module(module_path)

            return module

    @classmethod
    def load_pair(
        cls,
        pair: Pair,
        init_args: InitArgs = None,
    ) -> Optional[object]:
        module = cls.load_module(module_path=pair.module_path)
        instance = cls.load_instance(
            module=module,
            class_name=pair.class_name,
            init_args=init_args,
        )

        return instance

    @classmethod
    def load_module_list(
        cls,
        module_path_list: List[str],
    ) -> Optional[object]:
        """Load the first valid module among a `module_path_list`.

        Args:
            module_path_list (List[str]):
                List of `[module_path, class_name]` pair.

        Returns:
            Optional[ModuleType]:
                Loaded module or None.
        """

        module = None
        for module_path in module_path_list:
            module = cls.load_module(module_path=module_path)
            if module is not None:
                break

        return module

    @classmethod
    def load_pair_list(
        cls,
        pair_list: List[Pair],
        init_args: InitArgs = None,
    ) -> Optional[object]:
        """Load the first valid instance among a `pair_list`.

        Args:
            pairs (List[Pair]):
                List of Pair.

        Returns:
            Optional[ModuleType]:
                Loaded instance or None.
        """

        instance = None
        for pair in pair_list:
            instance = cls.load_pair(pair=pair, init_args=init_args)
            if instance is not None:
                break

        return instance



import datetime
import logging
import json
import pandas as pd

from degiro_connector.quotecast.models.metrics_storage import MetricsStorage
from degiro_connector.quotecast.models.quotecast_pb2 import Quotecast, Ticker
from typing import Dict, List, Union


class QuotecastParser:
    """Handle the payload returned from this endpoint :

    "https://degiro.quotecast.vwdservices.com/CORS/{session_id}"

    OVERALL EXPLANATION
    The endpoint aforementioned returns financial data measurement
    such as :
        * LastPrice of a Stock
        * LastVolume of a Stock
        * ...

    This endpoint use HTTP 1.1 protocol.

    Since there are no way to retrieve data asynchronously from
    HTTP 1.1, this endpoint uses Long-Polling method to retrieve data
    as soon as they are available.

    Thus to consume this endpoint one needs to recall it as soon as a
    response has been received.

    If no data updates are available a HEARTBEAT signal is sent every
    5 seconds.

    The session timeout is approximativaly 15 seconds.

    PAYLOAD DATA DESCRIPTION
    The payload sent from this endpoint can contain three kind of
    objects :
        * DATA
        * HEARTBEAT
        * MATCHING

    A DATA is an object with two elements :
        * REFERENCE
        * VALUE

    A HEARTBEAT is signal sent every 5 seconds if no data updates
    are available.

    A MATCHING is an object with three elements :
        * REFERENCE
        * VWD_ID
        * LABEL

    The REFERENCE is a unique identifier in Degiro's Quotecast API
    which refers to the financial data.

    The VALUE is the value of the financial data.

    THE VWD_ID is the id of the product (etf, option, stock, warrant...)
    from which we retrieve measurements (LastPrice, LastVolume...).

    The LABEL is the name of the measurements that we retrieve.

    The MATCHING table is only sent at the first call of this endpoint.

    PAYLOAD DATA STRUCTURE
    The payload is a list of messages serialized in JSON, as follows :
    ```json
    [
        MESSAGE_1,
        MESSAGE_2,
        ...
        MESSAGE_X
    ]
    ```

    What is called a MESSAGE here above, is an object with the two
    attributes :
        * "m"
        * "v"

    We can denote a message like this :
    ```python
    MESSAGE = {
        'm': MESSAGE_TYPE
        'v': [CODE1, CODE2]
    }
    ```

    Here we have the following properties:
        * m contains a MESSAGE_TYPE
        * MESSAGE_TYPE is a str
        * v is a list
        * v contains 2 elements
        * CODE1 is a str or an int
        * CODE2 is a float

    The MESSAGE_TYPE indicate the type of message, it can take the
    following __values :
        * "a_req" : subscription
        * "a_rel" : unsubscription
        * "d" : rejected subscription
        * "h" : heartbeat
        * "sr" : session invalid
        * "ue" : data not available
        * "un" : numeric data
        * "us" : string data
        * ... (this list may not be exhaustive)

    Depending on the MESSAGE_TYPE different kind of information are stored
    inside :
        * CODE1
        * CODE2

    If MESSAGE_TYPE = "a_req" or "a_rel" :
        * CODE1 : contains the product's VWD_ID and the PARAMETER_NAME.
        * CODE2 : contains the REFERENCE for the parameter in CODE1.
        * Example of MESSAGE :
        ```json
        {
            "m": "a_req",
            "v": ["365004197.B10Volume", 624239]
        }
        ```
        "365004197.B10Volume" <=> CODE1
        365004197 <=> VWD_ID
        B10Volume <=> PARAMETER_NAME
        624239 <=> REFERENCE

    If MESSAGE_TYPE = "un" or "us :
        * CODE1 : contains the reference number.
        * CODE2 : contains the value of the information referenced by
        CODE1.
        * Example of MESSAGE :
        ```json
        {
            "m": "a_req",
            "v": [624239, 115.85]
        }
        ```
        624239 <=> CODE1
        115.85 <=> CODE2
    """

    @classmethod
    def ticker_to_df(
        cls,
        ticker: Ticker,
        column_list: List[str] = None,
    ) -> pd.DataFrame:
        """Converts a ticker to a "pandas.DataFrame".
        Args:
            ticker (Ticker):
                Ticker to convert.
            column_list (List[str]):
                Mandatory fields : will be set to "None" if empty.
                Default to [].
        Returns:
            pandas.DataFrame:
                "pandas.DataFrame" containing the metrics.
                Each row depicts a specific product.
                Each column depicts a specific metric.
        """

        if column_list is None:
            column_list = list()

        ticker_dict = cls.ticker_to_dict(
            ticker=ticker,
            column_list=column_list,
        )

        df = pd.DataFrame(ticker_dict.values())

        return df

    @staticmethod
    def ticker_to_dict(
        ticker: Ticker,
        column_list: List[str] = None,
    ) -> Dict[str, Dict[str, Union[str, int]]]:
        # VWD_ID  # METRICS : NAME / VALUE
        """Converts a ticker to a "dict".
        Args:
            ticker (Ticker):
                Ticker to convert.
            column_list (List[str]):
                Mandatory fields : will be set to "None" if empty.
                Default to [].
        Returns:
            Dict[Union[str, int], Dict[str, Union[str, int]]]:
                Dict containing all the metrics grouped by "vwd_id".
        """

        if column_list is None:
            column_list = list()

        empty_list = [None] * len(column_list)
        empty_metrics = dict(zip(column_list, empty_list))
        empty_metrics[
            "response_datetime"
        ] = ticker.metadata.response_datetime.ToJsonString()
        empty_metrics["request_duration"] = (
            ticker.metadata.request_duration.ToMicroseconds() / 10**6
        )

        ticker_dict: Dict[str, Dict[str, Union[str, int]]] = dict()
        for product in ticker.products:
            ticker_dict[product] = empty_metrics.copy()
            ticker_dict[product]["vwd_id"] = product
            ticker_dict[product].update(ticker.products[product].metrics)

        return ticker_dict

    @staticmethod
    def build_ticker_from_quotecast(
        quotecast: Quotecast,
        references: Dict[int, List[str]] = None,
        ticker: Ticker = None,
    ) -> Ticker:
        """Build or update a Ticker metrics using a Quotecast object.

        Only the metrics which can be converted to float are supported.

        But that should be enough to handle all the real use cases.

        This was done to :
            * Keep the Ticker structure simple and light.
            * Have better performances during processing.

        Args:
            quotecast (Quotecast):
                Object containing the raw metrics.
            ticker (Ticker, optional):
                Object to update with the new metrics.
                Defaults to Ticker().
            references (Dict[int, List[str]], optional):
                The references dictionnary is a registry.
                It links the products :
                    * reference
                    * vwd_id
                    * metric
                Here is an example of how to populate it :
                    references[reference] = [vwd_id, metric]
                Defaults to dict().

        Raises:
            AttributeError:
                If the subscription is rejected.
                Or if the metric is unknown.

        Returns:
            Ticker: New or updated Ticker.
        """

        if references is None:
            references = dict()

        if ticker is None:
            ticker = Ticker()

        # SETUP PRODUCTS & METRICS
        message_array = json.loads(quotecast.json_data)
        for message in message_array:
            if message["m"] == "un":
                reference = message["v"][0]
                value = message["v"][1]
                product, metric = references[reference]
                ticker.products[product].metrics[metric] = value
            elif message["m"] == "us":
                reference = message["v"][0]
                value = message["v"][1]
                product, metric = references[reference]

                if value[4] == "-":
                    date = datetime.datetime.strptime(
                        value,
                        "%Y-%m-%d",
                    )
                    value = datetime.datetime.timestamp(date)
                    ticker.products[product].metrics[metric] = value
                elif value[2] == ":":
                    time = datetime.time.fromisoformat(value)
                    value = time.hour * 3600 + time.minute * 60 + time.second
                    ticker.products[product].metrics[metric] = value
                else:
                    # NOT CONVERTIBLE TO FLOAT
                    raise RuntimeWarning(
                        "Unsupported string metric : " f"{metric} = {message}"
                    )
            elif message["m"] == "a_req":
                references[message["v"][1]] = message["v"][0].rsplit(
                    sep=".",
                    maxsplit=1,
                )
            elif message["m"] == "a_rel":
                delete_list = []
                for reference in references:
                    if ".".join(references[reference]) == message["v"][0]:
                        delete_list.append(reference)

                for reference in delete_list:
                    del references[reference]
            elif message["m"] == "h":
                pass
            elif message["m"] == "ue":
                pass
            elif message["m"] == "d":
                raise AttributeError(f"Subscription rejected : {message}")
            else:
                raise AttributeError(f"Unknown metric : {message}")

        # SETUP PRODUCT LIST
        ticker.product_list.extend(ticker.products)

        # SETUP METADATA
        ticker.metadata.MergeFrom(quotecast.metadata)

        return ticker

    @property
    def references(self) -> Dict[int, List[str]]:
        return self.__references

    @property
    def ticker(self) -> Ticker:
        return self.__ticker

    @property
    def ticker_df(self) -> pd.DataFrame:
        ticker = self.__ticker
        ticker_df = self.ticker_to_df(ticker=ticker)
        return ticker_df

    @property
    def ticker_dict(
        self,
    ) -> Dict[str, Dict[str, Union[str, int]]]:
        # VWD_ID  # METRICS : NAME / VALUE
        ticker = self.__ticker
        ticker_dict = self.ticker_to_dict(ticker=ticker)
        return ticker_dict

    def __init__(self, forward_fill: bool = False):
        """
        Args:
            forward_fill (bool, optional):
                Whether or not we want to fill the new Ticker with
                previous received metrics.
                Default to False.
        """

        self.__forward_fill = forward_fill
        self.__metrics_storage = MetricsStorage()
        self.__references: Dict[int, List[str]] = dict()
        self.__ticker = Ticker()

        self.__logger = logging.getLogger(self.__module__)

    def put_quotecast(self, quotecast: Quotecast):
        forward_fill = self.__forward_fill
        metrics_storage = self.__metrics_storage
        references = self.__references

        ticker = self.build_ticker_from_quotecast(
            quotecast=quotecast,
            ticker=Ticker(),
            references=references,
        )

        if forward_fill is True:
            metrics_storage.fill_ticker(ticker=ticker)

        self.__ticker = ticker

    def rebuild_request(self) -> Quotecast.Request:
        """Rebuild the request from history (self.__references).

        Returns:
            Quotecast.Request:
                Request matching data-stream.
        """

        references = self.references
        request = Quotecast.Request()

        for vwd_id, metric in references.values():
            request.subscriptions[vwd_id].append(metric)

        return request


if __name__ == "__main__":
    data = (
        '[{"m":"h"},{"m":"a_req","v":["360015751.LastPrice",101]},'
        '{"m":"un","v":[101,119.900000]}]'
    )



# IMPORTATION STANDARD
from typing import Dict, List, Union

# IMPORTATION THIRD PARTY
# IMPORTATION INTERNAL
from degiro_connector.quotecast.models.quotecast_pb2 import Ticker


class MetricsStorage:
    @staticmethod
    def merge_tickers(
        ticker1: Ticker,
        ticker2: Ticker,
        update_only: bool = False,
    ):
        """Override metrics of ticker1 with ticker2's metrics.
        Args:
            ticker1 (Ticker): Ticker to fill.
            ticker2 (Ticker): Ticker used to fill.
            update_only (bool, optional):
                Whether or not we want to add products from "ticker2" to
                "ticker1".
        """

        if update_only is True:
            for ticker2_product in ticker2.products:
                if ticker2_product in ticker1.products:
                    ticker1.products[ticker2_product].metrics.update(
                        ticker2.products[ticker2_product].metrics
                    )
        else:
            for ticker2_product in ticker2.products:
                ticker1.products[ticker2_product].metrics.update(
                    ticker2.products[ticker2_product].metrics
                )

    @property
    def storage_ticker(self) -> Ticker:
        return self.__storage_ticker

    def __init__(self):
        self.__storage_ticker = Ticker()

    def add_metrics(self, ticker: Ticker):
        storage_ticker = self.__storage_ticker
        self.merge_tickers(
            ticker1=storage_ticker,
            ticker2=ticker,
            update_only=False,
        )

    def fill_ticker(self, ticker: Ticker):
        storage_ticker = self.__storage_ticker

        self.add_metrics(ticker=ticker)

        self.merge_tickers(
            ticker1=ticker,
            ticker2=storage_ticker,
            update_only=True,
        )



# IMPORTATION STANDARD
import logging
import requests
import time
from threading import Event

# IMPORTATION THIRD PARTY
from wrapt.decorators import synchronized

# IMPORTATION INTERNAL


class ModelConnection:
    @property
    def connected(self) -> Event:
        return self.__connected

    @property
    def timeout(self) -> int:
        return self.__timeout

    @property  # type: ignore
    @synchronized
    def session_id(self) -> str:
        if not self.__session_id:
            raise ConnectionError("Connection required.")

        if self.is_timeout_expired():
            self.__connected.clear()
            raise TimeoutError("Connection has probably expired.")

        return self.__session_id

    @session_id.setter  # type: ignore
    @synchronized
    def session_id(self, session_id: str):
        if session_id:
            self.__session_id = session_id
            self.__connected.set()
        else:
            self.__session_id = session_id
            self.__connected.clear()

    def __init__(
        self,
        timeout: int,  # quotecast : 15s / trading: 1800s
    ):
        self.__timeout = timeout

        self.__connected = Event()
        self.__last_success = 0
        self.__logger = logging.getLogger(self.__module__)
        self.__session_id = ""

    @synchronized
    def is_timeout_expired(self):
        if not self.__last_success:
            return False

        return (time.monotonic() - self.__last_success) > self.__timeout

    @synchronized
    def response_hook(self, response, *args, **kwargs):
        """This hook will intercept all the "requests.Response"."""

        timestamp = time.monotonic()
        status_code = response.status_code

        if self.__last_success < timestamp and status_code == 200:
            self.__last_success = timestamp

    def build_hooks(self):
        return {"response": [self.response_hook]}

    def setup_hooks(self, session: requests.Session):
        session.hooks.update(self.build_hooks())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: degiro_connector/quotecast/models/quotecast.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1degiro_connector/quotecast/models/quotecast.proto\x12\x1a\x64\x65giro_connector.quotecast\x1a\x1egoogle/protobuf/duration.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"v\n\x08Metadata\x12\x35\n\x11response_datetime\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x33\n\x10request_duration\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration\"\xbe\x03\n\tQuotecast\x12\x11\n\tjson_data\x18\x01 \x01(\t\x12\x36\n\x08metadata\x18\x02 \x01(\x0b\x32$.degiro_connector.quotecast.Metadata\x1a\xe5\x02\n\x07Request\x12W\n\rsubscriptions\x18\x01 \x03(\x0b\x32@.degiro_connector.quotecast.Quotecast.Request.SubscriptionsEntry\x12[\n\x0funsubscriptions\x18\x02 \x03(\x0b\x32\x42.degiro_connector.quotecast.Quotecast.Request.UnsubscriptionsEntry\x1aP\n\x12SubscriptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValue:\x02\x38\x01\x1aR\n\x14UnsubscriptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValue:\x02\x38\x01\"\xfd\x02\n\x06Ticker\x12\x36\n\x08metadata\x18\x01 \x01(\x0b\x32$.degiro_connector.quotecast.Metadata\x12\x42\n\x08products\x18\x02 \x03(\x0b\x32\x30.degiro_connector.quotecast.Ticker.ProductsEntry\x12\x14\n\x0cproduct_list\x18\x03 \x03(\t\x1a\x83\x01\n\x07Metrics\x12H\n\x07metrics\x18\x01 \x03(\x0b\x32\x37.degiro_connector.quotecast.Ticker.Metrics.MetricsEntry\x1a.\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\x1a[\n\rProductsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x39\n\x05value\x18\x02 \x01(\x0b\x32*.degiro_connector.quotecast.Ticker.Metrics:\x02\x38\x01\"\x85\x06\n\x05\x43hart\x12\x11\n\trequestid\x18\x01 \x01(\t\x12\r\n\x05start\x18\x02 \x01(\t\x12\x0b\n\x03\x65nd\x18\x03 \x01(\t\x12\x12\n\nresolution\x18\x04 \x01(\t\x12\x37\n\x06series\x18\x05 \x03(\x0b\x32\'.degiro_connector.quotecast.Chart.Serie\x1a\xc1\x02\n\x07Request\x12\x11\n\trequestid\x18\x01 \x01(\t\x12>\n\nresolution\x18\x02 \x01(\x0e\x32*.degiro_connector.quotecast.Chart.Interval\x12\x0f\n\x07\x63ulture\x18\x03 \x01(\t\x12:\n\x06period\x18\x04 \x01(\x0e\x32*.degiro_connector.quotecast.Chart.Interval\x12\x0e\n\x06series\x18\x05 \x03(\t\x12\n\n\x02tz\x18\x06 \x01(\t\x12I\n\x08override\x18\x07 \x03(\x0b\x32\x37.degiro_connector.quotecast.Chart.Request.OverrideEntry\x1a/\n\rOverrideEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1ak\n\x05Serie\x12\r\n\x05times\x18\x01 \x01(\t\x12\x0f\n\x07\x65xpires\x18\x02 \x01(\t\x12(\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.ListValue\x12\n\n\x02id\x18\x04 \x01(\t\x12\x0c\n\x04type\x18\x05 \x01(\t\"\xce\x01\n\x08Interval\x12\x08\n\x04PT1S\x10\x00\x12\t\n\x05PT15S\x10\x01\x12\t\n\x05PT30S\x10\x02\x12\x08\n\x04PT1M\x10\x03\x12\x08\n\x04PT5M\x10\x04\x12\t\n\x05PT15M\x10\x05\x12\t\n\x05PT30M\x10\x06\x12\t\n\x05PT60M\x10\x07\x12\x08\n\x04PT1H\x10\x08\x12\x07\n\x03P1D\x10\t\x12\x07\n\x03P1W\x10\n\x12\x07\n\x03P1M\x10\x0b\x12\x07\n\x03P3M\x10\x0c\x12\x07\n\x03P6M\x10\r\x12\x07\n\x03P1Y\x10\x0e\x12\x07\n\x03P3Y\x10\x0f\x12\x07\n\x03P5Y\x10\x10\x12\x08\n\x04P10Y\x10\x11\x12\x08\n\x04P50Y\x10\x12\x12\x07\n\x03YTD\x10\x13\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'degiro_connector.quotecast.models.quotecast_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _QUOTECAST_REQUEST_SUBSCRIPTIONSENTRY._options = None
  _QUOTECAST_REQUEST_SUBSCRIPTIONSENTRY._serialized_options = b'8\001'
  _QUOTECAST_REQUEST_UNSUBSCRIPTIONSENTRY._options = None
  _QUOTECAST_REQUEST_UNSUBSCRIPTIONSENTRY._serialized_options = b'8\001'
  _TICKER_METRICS_METRICSENTRY._options = None
  _TICKER_METRICS_METRICSENTRY._serialized_options = b'8\001'
  _TICKER_PRODUCTSENTRY._options = None
  _TICKER_PRODUCTSENTRY._serialized_options = b'8\001'
  _CHART_REQUEST_OVERRIDEENTRY._options = None
  _CHART_REQUEST_OVERRIDEENTRY._serialized_options = b'8\001'
  _METADATA._serialized_start=176
  _METADATA._serialized_end=294
  _QUOTECAST._serialized_start=297
  _QUOTECAST._serialized_end=743
  _QUOTECAST_REQUEST._serialized_start=386
  _QUOTECAST_REQUEST._serialized_end=743
  _QUOTECAST_REQUEST_SUBSCRIPTIONSENTRY._serialized_start=579
  _QUOTECAST_REQUEST_SUBSCRIPTIONSENTRY._serialized_end=659
  _QUOTECAST_REQUEST_UNSUBSCRIPTIONSENTRY._serialized_start=661
  _QUOTECAST_REQUEST_UNSUBSCRIPTIONSENTRY._serialized_end=743
  _TICKER._serialized_start=746
  _TICKER._serialized_end=1127
  _TICKER_METRICS._serialized_start=903
  _TICKER_METRICS._serialized_end=1034
  _TICKER_METRICS_METRICSENTRY._serialized_start=988
  _TICKER_METRICS_METRICSENTRY._serialized_end=1034
  _TICKER_PRODUCTSENTRY._serialized_start=1036
  _TICKER_PRODUCTSENTRY._serialized_end=1127
  _CHART._serialized_start=1130
  _CHART._serialized_end=1903
  _CHART_REQUEST._serialized_start=1264
  _CHART_REQUEST._serialized_end=1585
  _CHART_REQUEST_OVERRIDEENTRY._serialized_start=1538
  _CHART_REQUEST_OVERRIDEENTRY._serialized_end=1585
  _CHART_SERIE._serialized_start=1587
  _CHART_SERIE._serialized_end=1694
  _CHART_INTERVAL._serialized_start=1697
  _CHART_INTERVAL._serialized_end=1903
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: degiro_connector/quotecast/models/quotecast_relay.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from degiro_connector.quotecast.models import quotecast_pb2 as degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7degiro_connector/quotecast/models/quotecast_relay.proto\x12 degiro_connector.quotecast_relay\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x31\x64\x65giro_connector/quotecast/models/quotecast.proto\"5\n\tSetConfig\x12\x12\n\nuser_token\x18\x01 \x01(\x05\x12\x14\n\x0c\x61uto_connect\x18\x02 \x01(\x08\x32\xb2\x03\n\x0eQuotecastRelay\x12W\n\nset_config\x12+.degiro_connector.quotecast_relay.SetConfig\x1a\x1a.google.protobuf.BoolValue\"\x00\x12\x41\n\x07\x63onnect\x12\x16.google.protobuf.Empty\x1a\x1c.google.protobuf.StringValue\"\x00\x12M\n\nfetch_data\x12\x16.google.protobuf.Empty\x1a%.degiro_connector.quotecast.Quotecast\"\x00\x12[\n\tget_chart\x12).degiro_connector.quotecast.Chart.Request\x1a!.degiro_connector.quotecast.Chart\"\x00\x12X\n\tsubscribe\x12-.degiro_connector.quotecast.Quotecast.Request\x1a\x1a.google.protobuf.BoolValue\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'degiro_connector.quotecast.models.quotecast_relay_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SETCONFIG._serialized_start=205
  _SETCONFIG._serialized_end=258
  _QUOTECASTRELAY._serialized_start=261
  _QUOTECASTRELAY._serialized_end=695
# @@protoc_insertion_point(module_scope)



# IMPORTATION STANDARD
import logging
from typing import Dict, Union

# IMPORTATION THIRD PARTY
import requests
from google.protobuf import json_format

# IMPORTATION INTERNAL
from degiro_connector.core.constants import urls
from degiro_connector.core.abstracts.abstract_action import AbstractAction
from degiro_connector.trading.models.trading_pb2 import (
    Credentials,
    ProductSearch,
)


class ActionProductSearch(AbstractAction):
    PRODUCT_SEARCH_REQUEST_URL_MATCHING = {
        ProductSearch.RequestBonds.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_BONDS,
        ProductSearch.RequestETFs.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_ETFS,
        ProductSearch.RequestFunds.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_FUNDS,
        ProductSearch.RequestFutures.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_FUTURES,
        ProductSearch.RequestLeverageds.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_LEVERAGEDS,
        ProductSearch.RequestLookup.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_LOOKUP,
        ProductSearch.RequestOptions.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_OPTIONS,
        ProductSearch.RequestStocks.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_STOCKS,
        ProductSearch.RequestWarrants.DESCRIPTOR.full_name: urls.PRODUCT_SEARCH_WARRANTS,
    }

    @staticmethod
    def product_search_request_to_api(
        request: Union[
            ProductSearch.RequestBonds,
            ProductSearch.RequestETFs,
            ProductSearch.RequestFunds,
            ProductSearch.RequestFutures,
            ProductSearch.RequestLeverageds,
            ProductSearch.RequestLookup,
            ProductSearch.RequestOptions,
            ProductSearch.RequestStocks,
            ProductSearch.RequestWarrants,
        ],
    ) -> dict:
        request_dict = json_format.MessageToDict(
            message=request,
            including_default_value_fields=False,
            preserving_proto_field_name=False,
            use_integers_for_enums=True,
            descriptor_pool=None,
            float_precision=None,
        )

        return request_dict

    @staticmethod
    def product_search_to_grpc(payload: dict) -> ProductSearch:
        product_search = ProductSearch()
        product_search.response_datetime.GetCurrentTime()
        json_format.ParseDict(
            js_dict=payload,
            message=product_search,
            ignore_unknown_fields=True,
            descriptor_pool=None,
        )

        return product_search

    @classmethod
    def product_search(
        cls,
        request: Union[
            ProductSearch.RequestBonds,
            ProductSearch.RequestETFs,
            ProductSearch.RequestFunds,
            ProductSearch.RequestFutures,
            ProductSearch.RequestLeverageds,
            ProductSearch.RequestLookup,
            ProductSearch.RequestOptions,
            ProductSearch.RequestStocks,
            ProductSearch.RequestWarrants,
        ],
        session_id: str,
        credentials: Credentials,
        raw: bool = False,
        session: requests.Session = None,
        logger: logging.Logger = None,
    ) -> Union[ProductSearch, Dict, None]:
        """Search products.
        Args:
            request (StockList.Request):
                List of options that we want to retrieve from the endpoint.
                Example 1:
                    request = ProductSearch.RequestLookup(
                        search_text='APPLE',
                        limit=10,
                        offset=0,
                    )
                Example 2:
                    request = ProductSearch.RequestStocks(
                        index_id=5,
                        is_in_us_green_list=False,
                        stock_country_id=886,
                        offset=0,
                        limit=100,
                        require_total=True,
                        sort_columns='name',
                        sort_types='asc',
                    )
            session_id (str):
                API's session id.
            credentials (Credentials):
                Credentials containing the parameter "int_account".
            raw (bool, optional):
                Whether are not we want the raw API response.
                Defaults to False.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Returns:
            ProductSearch: API response.
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        url = cls.PRODUCT_SEARCH_REQUEST_URL_MATCHING[request.DESCRIPTOR.full_name]

        params = cls.product_search_request_to_api(
            request=request,
        )

        if credentials.int_account > 0:
            params["intAccount"] = credentials.int_account

        params["sessionId"] = session_id

        http_request = requests.Request(method="GET", url=url, params=params)
        prepped = session.prepare_request(http_request)
        response_raw = None

        try:
            response_raw = session.send(prepped)
            response_raw.raise_for_status()
            response_dict = response_raw.json()

            if raw is True:
                return response_dict
            else:
                return cls.product_search_to_grpc(
                    payload=response_dict,
                )
        except requests.HTTPError as e:
            status_code = getattr(response_raw, "status_code", "No status_code found.")
            text = getattr(response_raw, "text", "No text found.")
            logger.fatal(status_code)
            logger.fatal(text)
            return None
        except Exception as e:
            logger.fatal(e)
            return None

    def call(
        self,
        request: Union[
            ProductSearch.RequestBonds,
            ProductSearch.RequestETFs,
            ProductSearch.RequestFunds,
            ProductSearch.RequestFutures,
            ProductSearch.RequestLeverageds,
            ProductSearch.RequestLookup,
            ProductSearch.RequestOptions,
            ProductSearch.RequestStocks,
            ProductSearch.RequestWarrants,
        ],
        raw: bool = False,
    ) -> Union[ProductSearch, Dict, None]:
        connection_storage = self.connection_storage
        session_id = connection_storage.session_id
        session = self.session_storage.session
        credentials = self.credentials
        logger = self.logger

        return self.product_search(
            request=request,
            session_id=session_id,
            credentials=credentials,
            raw=raw,
            session=session,
            logger=logger,
        )



import logging
import os
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any

import orjson as json
import requests

from degiro_connector.core.constants.urls import (
    QUOTECAST,
    QUOTECAST_VERSION,
)
from degiro_connector.core.constants.headers import HEADERS as HEADER_MAP
from degiro_connector.quotecast.models.ticker import Ticker, TickerRequest
from degiro_connector.quotecast.models.metric import MetricType


class TickerFetcher:
    @staticmethod
    def build_logger() -> logging.Logger:
        return logging.getLogger(__name__)

    @staticmethod
    def build_credentials(location: Path) -> dict[str, Any]:
        content = os.environ.get("DEGIRO_ACCOUNT") or location.read_text()
        config_dict = json.loads(content)  # pylint: disable=maybe-no-member

        return config_dict

    @staticmethod
    def build_session(
        headers: dict | None = None,
        hooks: dict | None = None,
    ) -> requests.Session:
        """Setup a "requests.Session" object.
        Args:
            headers (dict, optional):
                Headers to used for the Session.
                Defaults to None.
            hooks (dict, optional):
                Hooks for the Session.
                Defaults to None.

        Returns:
            requests.Session:
                Session object with the right headers and hooks.
        """

        session = requests.Session()

        if isinstance(headers, dict):
            session.headers.update(headers)
        else:
            session.headers.update(HEADER_MAP)

        if isinstance(hooks, dict):
            session.hooks.update(hooks)

        return session

    @classmethod
    def get_session_id(
        cls,
        user_token: int,
        logger: logging.Logger | None = None,
        session: requests.Session | None = None,
    ) -> str | None:
        """Retrieves the "session_id" necessary to access the data-stream.
        Args:
            user_token (int):
                User identifier in Degiro's API.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Returns:
            str: Session id
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        url = QUOTECAST
        url = f"{url}/request_session"
        version = QUOTECAST_VERSION
        parameters = {"version": version, "userToken": user_token}
        data = '{"referrer":"https://trader.degiro.nl"}'
        request = requests.Request(method="POST", url=url, data=data, params=parameters)
        prepped = session.prepare_request(request=request)

        try:
            response = session.send(request=prepped)
            response_dict = json.loads(response.text)  # pylint: disable=maybe-no-member
        except Exception as e:
            logger.fatal(e)
            return None

        logger.info("get_session_id:response_dict: %s", response_dict)

        if "sessionId" in response_dict:
            return response_dict["sessionId"]
        else:
            return None

    @classmethod
    def fetch_ticker(
        cls,
        session_id: str,
        session: requests.Session | None = None,
        logger: logging.Logger | None = None,
    ) -> Ticker | None:
        """Fetches data from the feed.
        Args:
            session_id (str):
                API's session id.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Raises:
            BrokenPipeError:
                A new "session_id" is required.
        Returns:
            Quotecast:
                json_data : raw JSON data string.
                metadata.response_datetime : reception timestamp.
                metadata.request_duration : request duration.
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        url = f"{QUOTECAST}/{session_id}"
        request = requests.Request(method="GET", url=url)
        prepped = session.prepare_request(request=request)
        start_ns = time.perf_counter_ns()

        try:
            response = session.send(request=prepped)
            # We could have used : response.elapsed.total_seconds()
            duration_ns = time.perf_counter_ns() - start_ns

            if response.text == '[{"m":"sr"}]':
                raise BrokenPipeError('A new "session_id" is required.')

            ticker = Ticker(
                json_text=response.text,
                # There is no "date" header returned
                # We could have used : response.cookies._now
                response_datetime=datetime.now(),
                request_duration=timedelta(microseconds=duration_ns // 1000),
            )
        except Exception as e:
            logger.fatal(e)
            return None

        return ticker

    @staticmethod
    def build_ticker_request_payload(ticker_request: TickerRequest) -> str:
        """Build a payload like the following:
        '{"controlData":"a_req(360017018.LastDate);a_req(360017018.LastTime);a_req(360017018.LastPrice);"}'
        """

        request_map = ticker_request.request_map
        request_type = ticker_request.request_type
        request_function = "a_req" if (request_type == "subscription") else "a_rel"

        payload = '{"controlData":"'

        for product_id, metric_type_list in request_map.items():
            for metric_type in metric_type_list:
                if isinstance(metric_type, MetricType):
                    metric_name = metric_type.name
                else:
                    metric_name = metric_type
                payload += f"{request_function}({product_id}.{metric_name});"

        payload += '"}'

        return payload

    @classmethod
    def subscribe(
        cls,
        ticker_request: TickerRequest,
        session_id: str,
        session: requests.Session | None = None,
        logger: logging.Logger | None = None,
    ) -> bool | None:
        """Adds/removes metric from the data-stream.
        Args:
            ticker_request (TickerRequest):
                List of subscriptions & unsubscriptions to do.
                Example :
                    ticker_request: TickerRequest = TickerRequest(
                        request_type="subscription",
                        request_map={
                            "360017018" : [MetricType.LastDate, MetricType.LastPrice],
                            "AAPL.BATS,E" : [MetricType.LastPrice, MetricType.LastVolume],
                            "360015751" : [MetricType.LastPrice, MetricType.LastVolume],
                        },
                    )
            session_id (str):
                API's session id.
            session (requests.Session, optional):
                This object will be generated if None.
                Defaults to None.
            logger (logging.Logger, optional):
                This object will be generated if None.
                Defaults to None.
        Raises:
            BrokenPipeError:
                A new "session_id" is required.
        Returns:
            bool:
                Whether or not the subscription succeeded.
        """

        if logger is None:
            logger = cls.build_logger()
        if session is None:
            session = cls.build_session()

        url = QUOTECAST
        url = f"{url}/{session_id}"
        data = cls.build_ticker_request_payload(ticker_request=ticker_request)

        logger.info("subscribe:data %s", data[:100])

        session_request = requests.Request(method="POST", url=url, data=data)
        prepped = session.prepare_request(request=session_request)
        response_raw = None

        try:
            response_raw = session.send(request=prepped)
            response_raw.raise_for_status()

            if response_raw.text == '[{"m":"sr"}]':
                raise BrokenPipeError('A new "session_id" is required.')

            return True
        except Exception as e:
            logger.fatal(e)
            return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: degiro_connector/trading/models/trading_relay.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from degiro_connector.trading.models import trading_pb2 as degiro__connector_dot_trading_dot_models_dot_trading__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3degiro_connector/trading/models/trading_relay.proto\x12\x1e\x64\x65giro_connector.trading_relay\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a-degiro_connector/trading/models/trading.proto\"]\n\tSetConfig\x12:\n\x0b\x63redentials\x18\x01 \x01(\x0b\x32%.degiro_connector.trading.Credentials\x12\x14\n\x0c\x61uto_connect\x18\x02 \x01(\x08\"u\n\x0c\x43onfirmOrder\x12\x35\n\x0f\x63onfirmation_id\x18\x01 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12.\n\x05order\x18\x02 \x01(\x0b\x32\x1f.degiro_connector.trading.Order\"\xb3\x05\n\rProductSearch\x12\x45\n\x05\x62onds\x18\x01 \x01(\x0b\x32\x34.degiro_connector.trading.ProductSearch.RequestBondsH\x00\x12\x43\n\x04\x65tfs\x18\x02 \x01(\x0b\x32\x33.degiro_connector.trading.ProductSearch.RequestETFsH\x00\x12\x45\n\x05\x66unds\x18\x03 \x01(\x0b\x32\x34.degiro_connector.trading.ProductSearch.RequestFundsH\x00\x12I\n\x07\x66utures\x18\x04 \x01(\x0b\x32\x36.degiro_connector.trading.ProductSearch.RequestFuturesH\x00\x12O\n\nleverageds\x18\x05 \x01(\x0b\x32\x39.degiro_connector.trading.ProductSearch.RequestLeveragedsH\x00\x12G\n\x06lookup\x18\x06 \x01(\x0b\x32\x35.degiro_connector.trading.ProductSearch.RequestLookupH\x00\x12I\n\x07options\x18\x07 \x01(\x0b\x32\x36.degiro_connector.trading.ProductSearch.RequestOptionsH\x00\x12G\n\x06stocks\x18\x08 \x01(\x0b\x32\x35.degiro_connector.trading.ProductSearch.RequestStocksH\x00\x12K\n\x08warrants\x18\t \x01(\x0b\x32\x37.degiro_connector.trading.ProductSearch.RequestWarrantsH\x00\x42\t\n\x07request2\xcf\x13\n\x0cTradingRelay\x12u\n\rconfirm_order\x12,.degiro_connector.trading_relay.ConfirmOrder\x1a\x34.degiro_connector.trading.Order.ConfirmationResponse\"\x00\x12j\n\x0eproduct_search\x12-.degiro_connector.trading_relay.ProductSearch\x1a\'.degiro_connector.trading.ProductSearch\"\x00\x12U\n\nset_config\x12).degiro_connector.trading_relay.SetConfig\x1a\x1a.google.protobuf.BoolValue\"\x00\x12\x62\n\x0b\x63heck_order\x12\x1f.degiro_connector.trading.Order\x1a\x30.degiro_connector.trading.Order.CheckingResponse\"\x00\x12\x41\n\x07\x63onnect\x12\x16.google.protobuf.Empty\x1a\x1c.google.protobuf.StringValue\"\x00\x12J\n\x0c\x64\x65lete_order\x12\x1c.google.protobuf.StringValue\x1a\x1a.google.protobuf.BoolValue\"\x00\x12\x45\n\x10get_account_info\x12\x16.google.protobuf.Empty\x1a\x17.google.protobuf.Struct\"\x00\x12v\n\x14get_account_overview\x12\x31.degiro_connector.trading.AccountOverview.Request\x1a).degiro_connector.trading.AccountOverview\"\x00\x12Z\n\nget_agenda\x12(.degiro_connector.trading.Agenda.Request\x1a .degiro_connector.trading.Agenda\"\x00\x12}\n\x17get_cash_account_report\x12\x33.degiro_connector.trading.CashAccountReport.Request\x1a+.degiro_connector.trading.CashAccountReport\"\x00\x12G\n\x12get_client_details\x12\x16.google.protobuf.Empty\x1a\x17.google.protobuf.Struct\"\x00\x12_\n\x13get_company_profile\x12\x1c.google.protobuf.StringValue\x1a(.degiro_connector.trading.CompanyProfile\"\x00\x12]\n\x12get_company_ratios\x12\x1c.google.protobuf.StringValue\x1a\'.degiro_connector.trading.CompanyRatios\"\x00\x12?\n\nget_config\x12\x16.google.protobuf.Empty\x1a\x17.google.protobuf.Struct\"\x00\x12U\n\x13get_favourites_list\x12\x16.google.protobuf.Empty\x1a$.degiro_connector.trading.Favourites\"\x00\x12i\n\x18get_financial_statements\x12\x1c.google.protobuf.StringValue\x1a-.degiro_connector.trading.FinancialStatements\"\x00\x12g\n\x0fget_latest_news\x12,.degiro_connector.trading.LatestNews.Request\x1a$.degiro_connector.trading.LatestNews\"\x00\x12q\n\x13get_news_by_company\x12/.degiro_connector.trading.NewsByCompany.Request\x1a\'.degiro_connector.trading.NewsByCompany\"\x00\x12p\n\x12get_orders_history\x12/.degiro_connector.trading.OrdersHistory.Request\x1a\'.degiro_connector.trading.OrdersHistory\"\x00\x12_\n\x13get_products_config\x12\x16.google.protobuf.Empty\x1a..degiro_connector.trading.ProductSearch.Config\"\x00\x12m\n\x11get_products_info\x12..degiro_connector.trading.ProductsInfo.Request\x1a&.degiro_connector.trading.ProductsInfo\"\x00\x12Z\n\x14get_top_news_preview\x12\x16.google.protobuf.Empty\x1a(.degiro_connector.trading.TopNewsPreview\"\x00\x12\x82\x01\n\x18get_transactions_history\x12\x35.degiro_connector.trading.TransactionsHistory.Request\x1a-.degiro_connector.trading.TransactionsHistory\"\x00\x12^\n\nget_update\x12,.degiro_connector.trading.Update.RequestList\x1a .degiro_connector.trading.Update\"\x00\x12>\n\x06logout\x12\x16.google.protobuf.Empty\x1a\x1a.google.protobuf.BoolValue\"\x00\x12M\n\x0cupdate_order\x12\x1f.degiro_connector.trading.Order\x1a\x1a.google.protobuf.BoolValue\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'degiro_connector.trading.models.trading_relay_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SETCONFIG._serialized_start=225
  _SETCONFIG._serialized_end=318
  _CONFIRMORDER._serialized_start=320
  _CONFIRMORDER._serialized_end=437
  _PRODUCTSEARCH._serialized_start=440
  _PRODUCTSEARCH._serialized_end=1131
  _TRADINGRELAY._serialized_start=1134
  _TRADINGRELAY._serialized_end=3645
# @@protoc_insertion_point(module_scope)



from copy import deepcopy

import polars as pl

from degiro_connector.quotecast.models.metric import Metric
from degiro_connector.quotecast.models.ticker import Ticker
from degiro_connector.quotecast.tools.ticker_to_metric_list import TickerToMetricList


class TickerToDF:
    @staticmethod
    def merge_metric_list(
        current_data: list[Metric],
        new_data: list[Metric],
    ) -> list[Metric]:
        current_data = deepcopy(current_data)
        metric_map = {
            f"{metric.product_id} {metric.metric_type.value}": metric
            for metric in current_data
        }

        for new_metric in new_data:
            map_key = f"{new_metric.product_id} {new_metric.metric_type.value}"
            metric_map[map_key] = new_metric

        return list(metric_map.values())

    @staticmethod
    def build_df(metric_list: list[Metric]) -> pl.DataFrame:
        df = pl.DataFrame(
            data=[metric.model_dump() for metric in metric_list],
            schema={
                "product_id": pl.Utf8,
                "metric_type": pl.Utf8,
                "value": pl.Utf8,
            },
        )

        df = df.pivot(index="product_id", columns="metric_type", values="value")

        price_column_list = list(
            filter(lambda column: column.endswith("Price"), df.columns)
        )
        volume_column_list = list(
            filter(lambda column: column.endswith("Volume"), df.columns)
        )
        order_column_list = list(
            filter(lambda column: column.endswith("Orders"), df.columns)
        )

        df = df.with_columns(
            (pl.col("LastDate") + " " + pl.col("LastTime")).alias("LastDatetime")
        )
        df = df.drop(["LastDate", "LastTime"])
        df = df.with_columns(
            pl.col("LastDatetime")
            .str.strptime(pl.Datetime, format="%Y-%m-%d %H:%M:%S")
            .alias("LastDatetime")
        )
        df = df.with_columns(
            pl.col("LastDatetime").dt.replace_time_zone("Europe/Paris")
        )
        df = df.with_columns(
            pl.col(column).cast(pl.Float64) for column in price_column_list
        )
        df = df.with_columns(
            pl.col(column).cast(pl.Int64) for column in volume_column_list
        )
        df = df.with_columns(
            pl.col(column).cast(pl.Int64) for column in order_column_list
        )

        df_utc = df.with_columns(
            pl.col("LastDatetime").dt.convert_time_zone("UTC").alias("LastDatetimeUTC")
        )
        df_utc = df_utc.with_columns(
            pl.col("LastDatetimeUTC").dt.replace_time_zone(None)
        )
        df_utc = df_utc.drop("LastDatetime")

        return df_utc

    def __init__(self) -> None:
        self.__last_df = None
        self.__last_metric_list = []
        self.__stored_request_duration_map = {}
        self.__stored_response_datetime_map = {}
        self.__stored_metric_list = []
        self.__ticker_to_metric_list = TickerToMetricList()

    @property
    def last_df(self) -> pl.DataFrame | None:
        return self.__last_df

    @property
    def last_metric_list(self) -> list[Metric]:
        return self.__last_metric_list

    @property
    def stored_metric_list(self) -> list[Metric]:
        return self.__stored_metric_list

    @property
    def ticker_to_metric_list(self) -> TickerToMetricList:
        return self.__ticker_to_metric_list

    def add_request_duration_column(
        self,
        df: pl.DataFrame,
        last_metric_list: list[Metric],
        ticker: Ticker,
    ) -> pl.DataFrame:
        request_duration_map = {
            metric.product_id: ticker.request_duration.total_seconds()
            for metric in last_metric_list
        }
        self.__stored_request_duration_map.update(request_duration_map)

        df = df.with_columns(
            request_duration_s=pl.col("product_id")
            .map_elements(lambda x: self.__stored_request_duration_map.get(x, None))
            .cast(pl.Float64)
        )

        return df

    def add_response_datetime_column(
        self,
        df: pl.DataFrame,
        last_metric_list: list[Metric],
        ticker: Ticker,
    ) -> pl.DataFrame:
        response_datetime = {
            metric.product_id: ticker.response_datetime for metric in last_metric_list
        }
        self.__stored_response_datetime_map.update(response_datetime)

        df = df.with_columns(
            response_datetime=pl.col("product_id").map_elements(
                lambda x: self.__stored_response_datetime_map.get(x, None)
            )
        )
        df = df.with_columns(
            pl.col("response_datetime").dt.replace_time_zone("Europe/Paris")
        )

        df = df.with_columns(
            pl.col("response_datetime")
            .dt.convert_time_zone("UTC")
            .alias("response_datetime_utc")
        )
        df = df.with_columns(pl.col("response_datetime_utc").dt.replace_time_zone(None))
        df = df.drop("response_datetime")

        return df

    def parse(self, ticker: Ticker) -> pl.DataFrame | None:
        stored_metric_list = self.__stored_metric_list
        ticker_to_metric_list = self.__ticker_to_metric_list

        if ticker.json_text != '[{"m":"h"}]':
            last_metric_list = ticker_to_metric_list.parse(ticker=ticker)
            stored_metric_list = self.merge_metric_list(
                current_data=stored_metric_list,
                new_data=last_metric_list,
            )

            df = self.build_df(metric_list=stored_metric_list)
            df = self.add_request_duration_column(
                df=df,
                last_metric_list=last_metric_list,
                ticker=ticker,
            )
            df = self.add_response_datetime_column(
                df=df,
                last_metric_list=last_metric_list,
                ticker=ticker,
            )
        else:
            last_metric_list = []
            df = None

        self.__last_df = df
        self.__last_metric_list = last_metric_list
        self.__stored_metric_list = stored_metric_list

        return df



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from degiro_connector.quotecast.models import quotecast_pb2 as degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2
from degiro_connector.quotecast.models import quotecast_relay_pb2 as degiro__connector_dot_quotecast_dot_models_dot_quotecast__relay__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


class QuotecastRelayStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.set_config = channel.unary_unary(
                '/degiro_connector.quotecast_relay.QuotecastRelay/set_config',
                request_serializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__relay__pb2.SetConfig.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )
        self.connect = channel.unary_unary(
                '/degiro_connector.quotecast_relay.QuotecastRelay/connect',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                )
        self.fetch_data = channel.unary_unary(
                '/degiro_connector.quotecast_relay.QuotecastRelay/fetch_data',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.FromString,
                )
        self.get_chart = channel.unary_unary(
                '/degiro_connector.quotecast_relay.QuotecastRelay/get_chart',
                request_serializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.Request.SerializeToString,
                response_deserializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.FromString,
                )
        self.subscribe = channel.unary_unary(
                '/degiro_connector.quotecast_relay.QuotecastRelay/subscribe',
                request_serializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.Request.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )


class QuotecastRelayServicer(object):
    """Missing associated documentation comment in .proto file."""

    def set_config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def connect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def fetch_data(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_chart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def subscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_QuotecastRelayServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'set_config': grpc.unary_unary_rpc_method_handler(
                    servicer.set_config,
                    request_deserializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__relay__pb2.SetConfig.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
            'connect': grpc.unary_unary_rpc_method_handler(
                    servicer.connect,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            ),
            'fetch_data': grpc.unary_unary_rpc_method_handler(
                    servicer.fetch_data,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.SerializeToString,
            ),
            'get_chart': grpc.unary_unary_rpc_method_handler(
                    servicer.get_chart,
                    request_deserializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.Request.FromString,
                    response_serializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.SerializeToString,
            ),
            'subscribe': grpc.unary_unary_rpc_method_handler(
                    servicer.subscribe,
                    request_deserializer=degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.Request.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'degiro_connector.quotecast_relay.QuotecastRelay', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class QuotecastRelay(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def set_config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.quotecast_relay.QuotecastRelay/set_config',
            degiro__connector_dot_quotecast_dot_models_dot_quotecast__relay__pb2.SetConfig.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def connect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.quotecast_relay.QuotecastRelay/connect',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def fetch_data(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.quotecast_relay.QuotecastRelay/fetch_data',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_chart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.quotecast_relay.QuotecastRelay/get_chart',
            degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.Request.SerializeToString,
            degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Chart.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/degiro_connector.quotecast_relay.QuotecastRelay/subscribe',
            degiro__connector_dot_quotecast_dot_models_dot_quotecast__pb2.Quotecast.Request.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import orjson as json


from degiro_connector.quotecast.models.message import (
    Message,
    MessageNumeric,
    MessageRegistration,
    MessageText,
    MessageUnregistration,
)
from degiro_connector.quotecast.models.metric import (
    Metric,
    MetricType,
)
from degiro_connector.quotecast.models.ticker import Ticker


class TickerToMetricList:
    @staticmethod
    def from_ticker_to_message_list(ticker: Ticker) -> list[Message]:
        json_text = ticker.json_text
        message_list_raw = json.loads(json_text)  # pylint: disable=no-member
        message_list: list[Message] = []

        for message_raw in message_list_raw:
            if message_raw["m"] == "un":
                message_list.append(
                    MessageNumeric(
                        reference=message_raw["v"][0],
                        value=message_raw["v"][1],
                    ),
                )
            elif message_raw["m"] == "us":
                message_list.append(
                    MessageText(
                        reference=message_raw["v"][0],
                        value=message_raw["v"][1],
                    ),
                )
            elif message_raw["m"] == "a_req":
                message_list.append(
                    MessageRegistration(
                        metric_name=message_raw["v"][0],
                        reference=message_raw["v"][1],
                    ),
                )
            elif message_raw["m"] == "a_rel":
                message_list.append(
                    MessageUnregistration(
                        metric_name=message_raw["v"][0],
                        reference=message_raw["v"][1],
                    ),
                )
            elif message_raw["m"] == "h":
                pass
            elif message_raw["m"] == "ue":
                pass
            elif message_raw["m"] == "d":
                raise AttributeError(f"Subscription rejected : {message_raw}")
            else:
                raise AttributeError(f"Unknown metric : {message_raw}")

        return message_list

    def __init__(
        self,
        reference_map: dict[int, list] | None = None,
    ) -> None:
        """

        Parameters
        ----------
            reference_map: dict[int, list] | None
                Dictionnary storing the references returned by Degiro's Quotecast.
                Each reference number matches with a specific product/metric_type set.
                Example : {reference_number: [product_id, metric_type]}
        """
        # {reference: [product_id, metric_type]}
        self._reference_map: dict[int, list] = reference_map or {}

    def from_message_list_to_metric_list(
        self, message_list: list[Message]
    ) -> list[Metric]:
        reference_map = self._reference_map
        metric_list = []

        for message in message_list:
            if isinstance(message, MessageRegistration):
                reference_map[message.reference] = message.metric_name.rsplit(
                    sep=".", maxsplit=1
                )
            elif isinstance(message, MessageUnregistration):
                del reference_map[
                    message.reference
                ]  # crashes on purpose to detect inconsistency
            elif isinstance(message, (MessageNumeric, MessageText)):
                product_id, metric_type = reference_map[message.reference]
                metric_list.append(
                    Metric(
                        product_id=product_id,
                        metric_type=MetricType(metric_type),
                        value=message.value,
                    ),
                )

        return metric_list

    def parse(self, ticker: Ticker) -> list[Metric]:
        message_list = self.from_ticker_to_message_list(ticker=ticker)
        metric_list = self.from_message_list_to_metric_list(message_list=message_list)
        return metric_list



from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from degiro_connector.quotecast.models import quotecast_pb2 as _quotecast_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class SetConfig(_message.Message):
    __slots__ = ["auto_connect", "user_token"]
    AUTO_CONNECT_FIELD_NUMBER: _ClassVar[int]
    USER_TOKEN_FIELD_NUMBER: _ClassVar[int]
    auto_connect: bool
    user_token: int
    def __init__(self, user_token: _Optional[int] = ..., auto_connect: bool = ...) -> None: ...
